#![doc = include_str!("../README.md")]

use std::{
    fmt,
    fs::{read_dir, File},
    io::{Read, Write},
    path::PathBuf,
    process,
    str::FromStr,
};

use regex::Regex;
use rustyline::{error::ReadlineError, history::FileHistory, DefaultEditor, Editor};
use thiserror::Error;

#[derive(Debug, Error)]
#[error("{0}")]
struct StrError(&'static str);

#[derive(Debug, Error)]
enum SemVerParseError {
    #[error("Missing Major Version")]
    MissingMajor,

    #[error("Missing Minor Version")]
    MissingMinor,

    #[error("Missing Path Version")]
    MissingPatch,

    #[error("Too many dots")]
    TooManyParts,

    #[error("{0} is not a valid number")]
    NonNumeric(&'static str),
}

struct SemVer {
    major: usize,
    minor: usize,
    patch: usize,
}

impl FromStr for SemVer {
    type Err = SemVerParseError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut iter = s.split(".");
        let major = iter
            .next()
            .ok_or(SemVerParseError::MissingMajor)?
            .parse()
            .or(Err(SemVerParseError::NonNumeric("Major")))?;
        let minor = iter
            .next()
            .ok_or(SemVerParseError::MissingMinor)?
            .parse()
            .or(Err(SemVerParseError::NonNumeric("Minor")))?;
        let patch = iter
            .next()
            .ok_or(SemVerParseError::MissingPatch)?
            .parse()
            .or(Err(SemVerParseError::NonNumeric("Patch")))?;

        if iter.next().is_some() {
            Err(SemVerParseError::TooManyParts)
        } else {
            Ok(Self {
                major,
                minor,
                patch,
            })
        }
    }
}

impl fmt::Display for SemVer {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "\"{}.{}.{}\"", self.major, self.minor, self.patch)
    }
}

fn search(results: &mut Vec<PathBuf>, search_dir: &PathBuf) {
    if search_dir.is_dir() {
        for entry in read_dir(search_dir).unwrap() {
            let entry = entry.unwrap();
            let path = entry.path();

            if path.is_file() {
                let file_name = path.file_name().unwrap();
                if file_name == "Cargo.toml"
                    || file_name == "package.json"
                    || file_name == "tauri.conf.json"
                {
                    log::trace!("Found path: {path:?}");
                    results.push(path);
                }
            } else if path.is_dir() {
                let filename = path.file_name().unwrap();
                if filename != "target"
                    && filename != "dist"
                    && filename != "node_modules"
                    && filename != ".git"
                {
                    log::trace!("Searching directory: {filename:?}");
                    search(results, &path);
                }
            }
        }
    }
}

fn handle_input(input: rustyline::Result<String>) -> String {
    match input {
        Ok(line) => line,
        Err(ReadlineError::Interrupted | ReadlineError::Eof) => {
            process::exit(0);
        },
        Err(err) => {
            log::error!("Error: {:?}", err);
            process::exit(0);
        },
    }
}

fn apply(
    regex: &Regex,
    manifest: &PathBuf,
    rl: &mut Editor<(), FileHistory>,
) -> anyhow::Result<()> {
    let mut file_contents = String::new();
    File::open(manifest)?.read_to_string(&mut file_contents)?;

    if let Some(mat) = regex.find(&file_contents) {
        let (start, end) = file_contents.split_at(mat.start() - 1);
        let (prev_version, end) = end.split_at(mat.len() + 2);

        let new_version = handle_input(rl.readline(&format!(
            "Update: {:?} (Current: {})? (Leave Blank to Skip): ",
            manifest,
            prev_version.trim_matches('"'),
        )));
        let new_version = new_version.to_string();
        let should_update = !new_version.trim().is_empty();

        if let (true, Some(new_version)) = (should_update, new_version.parse::<SemVer>().ok()) {
            let mut write = File::create(manifest)?;

            write.write_all(start.as_bytes())?;
            write.write_all(new_version.to_string().as_bytes())?;
            write.write_all(end.as_bytes())?;
            write.flush()?;

            log::info!("Updated {manifest:?} successfully");
        } else if should_update {
            log::error!("Invalid SemVer ({new_version}). Skipping {manifest:?}");
        } else {
            log::info!("Skipping {manifest:?}");
        }

        Ok(())
    } else {
        Err(StrError("Cannot find version number in file").into())
    }
}

fn main() -> anyhow::Result<()> {
    simple_logger::SimpleLogger::new()
        .with_level(log::LevelFilter::Info)
        .with_module_level("rustyline", log::LevelFilter::Warn)
        .with_colors(true)
        .without_timestamps()
        .init()
        .unwrap();

    let mut rl = DefaultEditor::new()?;

    log::info!("Searching for manifests...");
    let mut results = vec![];
    search(&mut results, &PathBuf::from_str(".")?);
    log::info!("Found {} manifests", results.len());

    let regex = Regex::new("\\d+\\.\\d+\\.\\d+").unwrap();
    for manifest in results.into_iter() {
        if let Err(err) = apply(&regex, &manifest, &mut rl) {
            log::error!("Failed to update manifest: {manifest:?}. {err}");
        }
    }

    Ok(())
}
